"""
Defines helper methods useful for setting up ports, launching servers, and handling `ngrok`
"""

import os
import socket
import threading
from flask import Flask, request, jsonify, abort, send_file, render_template
import threading
import pkg_resources
from distutils import dir_util
import gradio as gr
import time
import json
from gradio.tunneling import create_tunnel
import urllib.request
from shutil import copyfile
import requests
import sys
import csv
import copy
import logging

INITIAL_PORT_VALUE = int(os.getenv(
    'GRADIO_SERVER_PORT', "7860"))  # The http server will try to open on port 7860. If not available, 7861, 7862, etc.
TRY_NUM_PORTS = int(os.getenv(
    'GRADIO_NUM_PORTS', "100"))  # Number of ports to try before giving up and throwing an exception.
LOCALHOST_NAME = os.getenv(
    'GRADIO_SERVER_NAME', "127.0.0.1")
GRADIO_API_SERVER = "https://api.gradio.app/v1/tunnel-request"

STATIC_TEMPLATE_LIB = pkg_resources.resource_filename("gradio", "templates/")
STATIC_PATH_LIB = pkg_resources.resource_filename("gradio", "static/")

app = Flask(__name__,
    template_folder=STATIC_TEMPLATE_LIB,
    static_folder=None)  # TODO (aliabid94): replace with default static
# handler
app.app_globals = {}

# Hide Flask default message
cli = sys.modules['flask.cli']
cli.show_server_banner = lambda *x: None

def set_meta_tags(title, description, thumbnail):
    app.app_globals.update({
        "title": title,
        "description": description,
        "thumbnail": thumbnail
    })


def set_config(config):
    app.app_globals["config"] = config


def get_first_available_port(initial, final):
    """
    Gets the first open port in a specified range of port numbers
    :param initial: the initial value in the range of port numbers
    :param final: final (exclusive) value in the range of port numbers, should be greater than `initial`
    :return:
    """
    for port in range(initial, final):
        try:
            s = socket.socket()  # create a socket object
            s.bind((LOCALHOST_NAME, port))  # Bind to the port
            s.close()
            return port
        except OSError:
            pass
    raise OSError(
        "All ports from {} to {} are in use. Please close a port.".format(
            initial, final
        )
    )


@app.route("/", methods=["GET"])
def main():
    return render_template("index.html",
        title=app.app_globals["title"],
        description=app.app_globals["description"],
        thumbnail=app.app_globals["thumbnail"],
    )


@app.route("/static/<path:path>")
def static(path):
    return send_file(os.path.join(STATIC_PATH_LIB, path))


@app.route("/config/", methods=["GET"])
def config():
    return jsonify(app.app_globals["config"])


@app.route("/enable_sharing/<path:path>", methods=["GET"])
def enable_sharing(path):
    if path == "None":
        path = None
    app.app_globals["config"]["share_url"] = path
    return jsonify(success=True)
    

@app.route("/api/predict/", methods=["POST"])
def predict():
    raw_input = request.json["data"]
    prediction, durations = app.interface.process(raw_input)
    output = {"data": prediction, "durations": durations}
    return jsonify(output)


@app.route("/api/flag/", methods=["POST"])
def flag():
    flag_path = os.path.join(app.cwd, app.interface.flagging_dir)
    os.makedirs(flag_path,
                exist_ok=True)
    output = {'inputs': [app.interface.input_interfaces[
        i].rebuild(
        flag_path, request.json['data']['input_data'][i]) for i
        in range(len(app.interface.input_interfaces))],
        'outputs': [app.interface.output_interfaces[
            i].rebuild(
            flag_path, request.json['data']['output_data'][i])
            for i
        in range(len(app.interface.output_interfaces))]}

    log_fp = "{}/log.csv".format(flag_path)

    is_new = not os.path.exists(log_fp)

    with open(log_fp, "a") as csvfile:
        headers = ["input_{}".format(i) for i in range(len(
            output["inputs"]))] + ["output_{}".format(i) for i in
                                    range(len(output["outputs"]))]
        writer = csv.DictWriter(csvfile, delimiter=',',
                                lineterminator='\n',
                                fieldnames=headers)
        if is_new:
            writer.writeheader()

        writer.writerow(
            dict(zip(headers, output["inputs"] +
                        output["outputs"]))
        )
        return jsonify(success=True)


@app.route("/api/interpret/", methods=["POST"])
def interpret():
    raw_input = request.json["data"]
    if app.interface.interpretation == "default":
        interpreter = gr.interpretation.default()
        processed_input = []
        for i, x in enumerate(raw_input):
            input_interface = copy.deepcopy(app.interface.input_interfaces[i])
            interface_type = type(input_interface)
            if interface_type in gr.interpretation.expected_types:
                input_interface.type = gr.interpretation.expected_types[interface_type]
            processed_input.append(input_interface.preprocess(x))
    else:
        processed_input = [input_interface.preprocess(raw_input[i])
                            for i, input_interface in enumerate(app.interface.input_interfaces)]
        interpreter = app.interface.interpretation
    interpretation = interpreter(app.interface, processed_input)
    return jsonify(interpretation)


@app.route("/file/<path:path>", methods=["GET"])
def file(path):
    return send_file(os.path.join(app.cwd, path))

def start_server(interface, server_port=None):
    if server_port is None:
        server_port = INITIAL_PORT_VALUE
    port = get_first_available_port(
        server_port, server_port + TRY_NUM_PORTS
    )
    app.interface = interface
    app.cwd = os.getcwd()
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)
    process = threading.Thread(target=app.run, kwargs={"port": port})
    process.start()
    return port, app, process


def close_server(process):
    process.terminate()
    process.join()

def url_request(url):
    try:
        req = urllib.request.Request(
            url=url, headers={"content-type": "application/json"}
        )
        res = urllib.request.urlopen(req, timeout=10)
        return res
    except Exception as e:
        raise RuntimeError(str(e))


def setup_tunnel(local_server_port):
    response = url_request(GRADIO_API_SERVER)
    if response and response.code == 200:
        try:
            payload = json.loads(response.read().decode("utf-8"))[0]
            return create_tunnel(payload, LOCALHOST_NAME, local_server_port)

        except Exception as e:
            raise RuntimeError(str(e))


def url_ok(url):
    try:
        r = requests.head(url)
        return r.status_code == 200
    except ConnectionError:
        return False



from gradio.inputs import Image, Textbox
from gradio.outputs import Label
from gradio import processing_utils
from skimage.segmentation import slic
import numpy as np

expected_types = {
    Image: "numpy",
}

def default(separator=" ", n_segments=20):
    """
    Basic "default" interpretation method that uses "leave-one-out" to explain predictions for
    the following inputs: Image, Text, and the following outputs: Label. In case of multiple
    inputs and outputs, uses the first component.
    """
    def tokenize_text(text):
        leave_one_out_tokens = []
        tokens = text.split(separator)
        for idx, _ in enumerate(tokens):
            new_token_array = tokens.copy()
            del new_token_array[idx]
            leave_one_out_tokens.append(new_token_array)
        return leave_one_out_tokens

    def tokenize_image(image):
        segments_slic = slic(image, n_segments=20, compactness=10, sigma=1)
        leave_one_out_tokens = []
        replace_color = np.mean(image, axis=(0, 1))
        for (i, segVal) in enumerate(np.unique(segments_slic)):
            mask = segments_slic == segVal
            white_screen = np.copy(image)
            white_screen[segments_slic == segVal] = replace_color
            leave_one_out_tokens.append((mask, white_screen))
        return leave_one_out_tokens

    def score_text(interface, leave_one_out_tokens, text):
        tokens = text.split(separator)
        original_output = interface.run_prediction([text])

        scores_by_words = []
        for idx, input_text in enumerate(leave_one_out_tokens):
            perturbed_text = separator.join(input_text)
            perturbed_output = interface.run_prediction([perturbed_text])
            score = quantify_difference_in_label(interface, original_output, perturbed_output)
            scores_by_words.append(score)

        scores_by_char = []
        for idx, token in enumerate(tokens):
            if idx != 0:
                scores_by_char.append((" ", 0))
            for char in token:
                scores_by_char.append((char, scores_by_words[idx]))
        
        return scores_by_char

    def score_image(interface, leave_one_out_tokens, image):
        output_scores = np.zeros((image.shape[0], image.shape[1]))
        original_output = interface.run_prediction([image])

        for mask, perturbed_image in leave_one_out_tokens:
            perturbed_output = interface.run_prediction([perturbed_image])
            score = quantify_difference_in_label(interface, original_output, perturbed_output)
            output_scores += score * mask

        max_val, min_val = np.max(output_scores), np.min(output_scores)
        if max_val > 0:
            output_scores = (output_scores - min_val) / (max_val - min_val)
        return output_scores.tolist()

    def quantify_difference_in_label(interface, original_output, perturbed_output):
        post_original_output = interface.output_interfaces[0].postprocess(original_output[0])
        post_perturbed_output = interface.output_interfaces[0].postprocess(perturbed_output[0])
        original_label = post_original_output["label"]
        perturbed_label = post_perturbed_output["label"]

        # Handle different return types of Label interface
        if "confidences" in post_original_output:
            original_confidence = original_output[0][original_label]
            perturbed_confidence = perturbed_output[0][original_label]
            score = original_confidence - perturbed_confidence
        else:
            try:  # try computing numerical difference
                score = float(original_label) - float(perturbed_label)
            except ValueError:  # otherwise, look at strict difference in label
                score = int(not(perturbed_label == original_label))
        return score

    def default_interpretation(interface, x):
        if isinstance(interface.input_interfaces[0], Textbox) \
                and isinstance(interface.output_interfaces[0], Label):
            leave_one_out_tokens = tokenize_text(x[0])
            return [score_text(interface, leave_one_out_tokens, x[0])]
        if isinstance(interface.input_interfaces[0], Image) \
                and isinstance(interface.output_interfaces[0], Label):
            leave_one_out_tokens = tokenize_image(x[0])
            return [score_image(interface, leave_one_out_tokens, x[0])]
        else:
            print("Not valid input or output types for 'default' interpretation")

    return default_interpretation

